use axum::Router;
use tokio::{net::TcpListener, task::JoinHandle};

use crate::{config::ApiConfig, state::ApiState};

pub mod v0;

pub struct Api {
    server_thread: JoinHandle<()>,
}

impl Api {
    pub fn stop(&self) {
        self.server_thread.abort();
    }
}

pub struct ApiBuilder {
    bind_ip: Option<String>,
    port: Option<u16>,
    version: Option<String>,
}

impl Default for ApiBuilder {
    fn default() -> Self {
        ApiBuilder {
            bind_ip: None,
            port: None,
            version: None,
        }
    }
}

impl From<ApiConfig> for ApiBuilder {
    fn from(value: ApiConfig) -> Self {
        ApiBuilder {
            bind_ip: value.ip,
            port: value.port,
            version: value.version,
        }
    }
}

impl ApiBuilder {
    pub async fn build(&self, state: ApiState) -> Api {
        let mut app: Router = Router::new();
        match &self.version {
            Some(v) if v == "v0" => app = app.nest(&format!("/{}", v), v0::get_router(state)),
            _ => app = app.nest("/v0", v0::get_router(state)),
        }

        let ip = match &self.bind_ip {
            Some(addr) => addr,
            None => "127.0.0.1",
        };
        let port = match &self.port {
            Some(p) => p,
            None => &9981,
        };
        let listener = TcpListener::bind(&format!("{}:{}", ip, port))
            .await
            .unwrap();

        let task_handle = tokio::spawn({
            async move {
                axum::serve(listener, app).await.unwrap();
            }
        });

        Api {
            server_thread: task_handle,
        }
    }
}
