use std::sync::Arc;

use chrono::Utc;
use cozo::DbInstance;
use jsonwebtoken::{DecodingKey, EncodingKey, Validation};
use tracing::debug;

use crate::{
    api::v0::app::{AppDescription, AppId},
    comm::messages::MessageContent,
    state::{queries, types::ElementId},
};

use super::{
    types::{Element, ElementContent},
    State,
};

pub struct ApiState {
    state: Arc<State>,
    jwt_encoding_key: EncodingKey,
    jwt_decoding_key: DecodingKey,
    jwt_validation: Validation,
}

impl ApiState {
    pub fn new(state: Arc<State>, jwt_secret: &str) -> Self {
        let mut validation = Validation::default();
        validation.set_required_spec_claims(&vec!["sub"]);
        ApiState {
            state: state,
            jwt_encoding_key: EncodingKey::from_secret(jwt_secret.as_bytes()),
            jwt_decoding_key: DecodingKey::from_secret(jwt_secret.as_bytes()),
            jwt_validation: validation,
        }
    }

    pub fn add_app(&self, description: &AppDescription) -> anyhow::Result<AppId> {
        let id = AppId::new();
        let last_access = Utc::now();
        queries::apps::add(
            self.db(),
            &id,
            &last_access,
            &description.name,
            &description.desc_text,
        )?;
        debug!("Successfully added app");

        Ok(id)
    }

    pub fn app_exists(&self, id: &AppId) -> anyhow::Result<bool> {
        queries::apps::exists(self.db(), id)
    }

    pub fn create_element(&self, content: &ElementContent) -> anyhow::Result<ElementId> {
        let id = ElementId::new();
        queries::elements::add(self.db(), &id, &content, None, true)?;
        debug!("Added element {{{}}}", &id.to_string());

        self.state.send_to_peers(MessageContent::CreateElement {
            id: id.clone(),
            content: content.clone(),
        });
        Ok(id)
    }

    pub fn write_element_content(
        &self,
        id: &ElementId,
        content: &ElementContent,
    ) -> anyhow::Result<()> {
        queries::elements::set_content(self.db(), id, content)?;
        queries::elements::set_local_changes(self.db(), id, true)?;
        debug!("Wrote element content {{{}}}", &id.to_string());

        Ok(())
    }

    pub fn remove_element(&self, id: &ElementId) -> anyhow::Result<()> {
        let res = self.state.remove_element(id);
        debug!("Removed element {{{}}}", &id.to_string());

        res
    }

    pub fn get_element(&self, id: &ElementId) -> anyhow::Result<Element> {
        self.state.get_element(id)
    }

    fn db(&self) -> &DbInstance {
        &self.state.db
    }

    pub fn jwt_encoding_key(&self) -> &EncodingKey {
        &self.jwt_encoding_key
    }

    pub fn jwt_decoding_key(&self) -> &DecodingKey {
        &self.jwt_decoding_key
    }

    pub fn jwt_validation(&self) -> &Validation {
        &self.jwt_validation
    }
}

#[cfg(test)]
mod tests {
    use super::ApiState;
    use crate::state::{types::ElementContent, State};

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_create() {
        tracing_subscriber::fmt().pretty().init();
        let state = ApiState::new(
            State::new().await.unwrap(),
            "abcdabcdabcdabcdabcdabcdabcdabcd",
        );
        let id = state
            .create_element(&ElementContent::Text("Test-text".to_string()))
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text".to_string()),
            el.content().to_owned()
        )
    }

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_write() {
        tracing_subscriber::fmt().pretty().init();
        let state = ApiState::new(
            State::new().await.unwrap(),
            "abcdabcdabcdabcdabcdabcdabcdabcd",
        );
        let id = state
            .create_element(&ElementContent::Text("Test-text".to_string()))
            .unwrap();
        state
            .write_element_content(&id, &ElementContent::Text("Test-text 2".to_string()))
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text 2".to_string()),
            el.content().to_owned()
        )
    }
}
