use std::sync::Arc;

use cozo::DbInstance;
use tracing::debug;

use crate::{comm::Peer, state::queries};

use super::{
    types::{Element, ElementContent, ElementId, MessageId, PeerId},
    State,
};

//TODO: Notify API about changes
pub struct CommState {
    state: Arc<State>,
}

impl CommState {
    pub fn new(state: Arc<State>) -> Self {
        CommState { state: state }
    }

    pub fn add_received_element(
        &self,
        id: &ElementId,
        content: &ElementContent,
        latest_message: &MessageId,
    ) -> anyhow::Result<()> {
        queries::elements::add(
            self.db(),
            &id,
            &content,
            Some(latest_message.to_owned()),
            false,
        )?;
        debug!("Added element {{{}}}", &id.to_string());

        Ok(())
    }

    pub fn update_element_content(
        &self,
        id: &ElementId,
        content: &ElementContent,
        latest_message: &MessageId,
    ) -> anyhow::Result<()> {
        //TODO: resolve potential conflicts with local changes
        queries::elements::set_content(self.db(), id, content)?;
        queries::elements::set_latest_message(self.db(), id, Some(latest_message.to_owned()))?;
        debug!("Updated element {{{}}}", &id.to_string());

        Ok(())
    }

    pub fn remove_element(&self, id: &ElementId) -> anyhow::Result<()> {
        let res = self.state.remove_element(id);
        debug!("Removed element {{{}}}", &id.to_string());

        res
    }

    pub fn get_element(&self, id: &ElementId) -> anyhow::Result<Element> {
        self.state.get_element(id)
    }

    pub fn set_peer(&self, id: &PeerId, name: &str) -> anyhow::Result<()> {
        queries::peers::put(self.db(), id, name)?;
        debug!("Set peer {{{}}}", id.to_string());

        Ok(())
    }

    pub fn get_peers(&self) -> anyhow::Result<Vec<Peer>> {
        self.state.get_peers()
    }

    fn db(&self) -> &DbInstance {
        &self.state.db
    }
}

#[cfg(test)]
mod tests {
    use super::CommState;
    use crate::state::{
        types::{ElementContent, ElementId, MessageId},
        State,
    };

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_add() {
        tracing_subscriber::fmt().pretty().init();
        let state = CommState::new(State::new().await.unwrap());
        let id = ElementId::new();
        state
            .add_received_element(
                &id,
                &ElementContent::Text("Test-text".to_string()),
                &MessageId::new(),
            )
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text".to_string()),
            el.content().to_owned()
        )
    }

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_update() {
        tracing_subscriber::fmt().pretty().init();
        let state = CommState::new(State::new().await.unwrap());
        let id = ElementId::new();
        state
            .add_received_element(
                &id,
                &ElementContent::Text("Test-text".to_string()),
                &MessageId::new(),
            )
            .unwrap();
        state
            .update_element_content(
                &id,
                &ElementContent::Text("Test-text 2".to_string()),
                &MessageId::new(),
            )
            .unwrap();
        let el = state.get_element(&id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text 2".to_string()),
            el.content().to_owned()
        )
    }
}
